(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                  projet Estime, INRIA-Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 2007-2011 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: parfun.ml,v 1.34.2.1 2011-09-21 13:08:03 fclement Exp $ *)

module Uo = Ucommand_option;;

exception Parfun_error;;

type ('x, 'a, 'b) worker = {
   io : IO.ic * IO.oc;
   f : 'a Sklstream.t -> 'b Sklstream.t;
};;

type master = { spawnd_io : Skl_comm.iochans; };;

type ('x, 'a, 'b) startup_ret =
   | Worker of ('x, 'a, 'b) worker
   | Master of master
;;

let request_spawnd io req =
  match
    Skl_comm.request_spawnd io req
  with
  | Some (Skl_comm.Success addr) ->
    addr
  | Some Skl_comm.Failure | None ->
    raise Parfun_error
;;

(* Core of the parallel implementation *)

let rec start_tree :
'a 'b. Skl_comm.iochans -> ('a, 'b) Skl.skel -> Iotree.ioatree =
  fun io tree ->
    match tree with
    | Sk.Skl.Par par ->
      Sk.with_par par {
        Sk.f_par = fun t1 t2 ->
          let addr_tree1 = start_tree io t1 in
          let addr_tree2 = start_tree io t2 in
          Iotree.IOABranch (addr_tree1, addr_tree2)
      }
    | Sk.Skl.Rails rails ->
      Sk.with_rails rails {
        Sk.f_rails = fun ska ->
          let addr_treev =
            Array.map (start_tree io) ska
          in
          Iotree.IOALeafs addr_treev
      }
    | Sk.Skl.Sum sum ->
      Sk.with_sum sum {
        Sk.f_sum = fun _ t1 t2 ->
          let addr_tree1 = start_tree io t1 in
          let addr_tree2 = start_tree io t2 in
          Iotree.IOASumSplit (addr_tree1, addr_tree2)
      }
    | Sk.Skl.Pipe pipe ->
      Sk.with_pipe pipe {
        Sk.f_pipe = fun t1 t2 ->
          let addr_tree1 = start_tree io t1 in
          let addr_tree2 = start_tree io t2 in
          let spawn_request =
            Skl_comm.Pipe_req
              (addr_tree1, addr_tree2) in
          let addr = request_spawnd io spawn_request in
          Iotree.IOALeaf addr
      }
    | Sk.Skl.Atom atom ->
      Sk.with_atom atom {
        Sk.f_atom = fun params llexp color ->
          let spawn_request =
            let (p, id) = Expor.get_exportable (params, llexp) in
            Skl_comm.Atom_req
              (p, id, color) in
          let addr = request_spawnd io spawn_request in
          Iotree.IOALeaf addr
      }
    | Sk.Skl.Id id_skl ->
      Sk.with_id id_skl {
        Sk.f_id = fun _ ->
          Iotree.IOALeafId
      }
    | Sk.Skl.Prod prod ->
      Sk.with_prod prod {
        Sk.f_prod = fun _ _ skl1 skl2 ->
          let addr_tree1 = start_tree io skl1 in
          let addr_tree2 = start_tree io skl2 in
          Iotree.IOASplit (addr_tree1, addr_tree2)
      }
    | Sk.Skl.Loop loop ->
      Sk.with_loop loop {
        Sk.f_loop = fun _ ptree tree ->
          let addr_tree1 = start_tree io ptree in
          let addr_tree2 = start_tree io tree in
          let spawn_request =
            Skl_comm.Loop_req
              (addr_tree1, addr_tree2) in
          let addr = request_spawnd io spawn_request in
          Iotree.IOALeaf addr
      }
;;

type parfun =
  { parfun : 'a 'b. ('a, 'b) Skl.skel -> 'a Sklstream.t -> 'b Sklstream.t; }
;;

let parfun io skl is =
  let tree =
    Iotree.iotree_of_ioatree (* Connects the tree *)
      (start_tree io skl) in
  match IO.fork () with
  | IO.Parent _ -> Iotree.receive_in_tree tree
  | IO.Child ->
    Iotree.send_in_tree tree is;
    exit 0
;;

type arg_master = {
   addr : Unix.sockaddr;
};;

type arg_worker = {
   f_id : int;
};;

type parsed_args =
   | Arg_master of arg_master
   | Arg_worker of arg_worker
;;

let get_usage () = "This is a sklml program";;

let parse_argv argv =
  let mode = ref None in
  let (mkmaster, mkslave) =
    (fun x ->
      mode := Some
        (Arg_master { addr = IO.read_sockaddr x; })),
    (fun x ->
      mode := Some
        (Arg_worker { f_id = x; })) in
  let arglist = Arg.align
    [ "-atom", Arg.Int mkslave, ""
    ; "-master", Arg.String mkmaster, ""
    ] in
  let to_uo (k, s, d) =
    { Uo.key = k; Uo.spec = s; Uo.doc = d; } in
  Uo.add_all (List.map to_uo arglist);
  let fail m =
    Log.say Log.Err "%s" m;
    exit 2 in
  try
    Uo.parse_argv argv ignore (get_usage ());
    match !mode with
    | Some m -> m
    | None -> fail "Mode unspecified, see help using -help option"
  with
    | Arg.Bad _ -> fail "Bad argument, see help using -help option"
    | Arg.Help m -> fail m
    | Failure m -> fail m
;;

let startup argv =
  match parse_argv argv with
  | Arg_master { addr = addr; } ->
    let io = IO.connect addr in
    Master { spawnd_io = io; }
  | Arg_worker { f_id = f_id; } ->
    let f_gen =
      (Expor.unsafe_retreive f_id) in
    let (fifo_in, fifo_out) =
      Skl_comm.open_skl_fifo false (Unix.getpid ()) in
    let params = IO.of_channel fifo_in in
    let f = f_gen params in
    let (sock, addr) = IO.create_bound_socket None in
    let port = IO.get_port addr in
    Unix.listen sock 1; (* listen before giving the port *)
    IO.to_channel fifo_out port true;
    let io, _ = IO.wait_connection sock true in
    Worker { io = io; f = f; }
;;
(** Startup function, if -atom N is spotted in [argv] then [startup
  argv] runs the atom code, otherwise, starts as master and connects
  to the spawnd daemon given as argument. *)

let run_worker w =
  let { io = (ic, oc); f = f; } = w in
  Sklstream.to_channel (f (Sklstream.of_channel ic)) oc
;;

let pardo f =
  match startup (Sys.argv) with
  | Master { spawnd_io = io; } ->
    f { parfun = fun x -> parfun io x; }
  | Worker w ->
    run_worker w;
    Log.say (Log.Log 0) "Worker job done."
;;
(** Entry point. *)
