(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*                  projet Estime, INRIA-Rocquencourt                  *)
(*                                                                     *)
(*  Copyright 2007-2011 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: slave.ml,v 1.26.2.1 2011-09-21 13:08:04 fclement Exp $ *)

type node = {
   color : int;
   link : bool;
   exe_path : string;
   local_addr : Unix.inet_addr;
   master : Unix.sockaddr;
};;

(* val to_tree : ('a, 'b) Iotree.ioatree -> 'a Sklstream.t -> 'b Sklstream.t *)
let to_tree tree is =
  let tree = Iotree.iotree_of_ioatree tree in
  match IO.fork () with
  | IO.Child ->
    Iotree.send_in_tree tree is;
    exit 0
  | IO.Parent _ ->
    Iotree.receive_in_tree tree
;;

(* val spawn_atom : string -> Spawnlib.closure_id -> Spawnlib.spawn_return *)
let spawn_atom skl_path (params, closure_id) =
  let (cmd, args) =
    (skl_path, [| skl_path; "-atom"; string_of_int closure_id; |]) in
  try
    match IO.fork () with
    | IO.Child -> Unix.execvp cmd args
    | IO.Parent pid ->
      let (fifo_in, fifo_out) = Skl_comm.open_skl_fifo true pid in
      IO.to_channel fifo_out params true;
      let port = IO.of_channel fifo_in in
      Skl_comm.clean_skl_fifo pid;
      Spawnd_types.Spawn_success
        { Spawnd_types.worker_port = port; }
  with
  | IO.IOError msg ->
    Log.say Log.Err "IO error is Slave.spawn_atom: %s" msg;
    Spawnd_types.Spawn_failure
  | End_of_file | Scanf.Scan_failure _ ->
    Log.say Log.Err "Scanf.bscanf error in Slave.spawn_atom";
    Spawnd_types.Spawn_failure
;;

(* val spawn_loop : ('a, bool) Iotree.ioatree -> ('a, 'a) Iotree.ioatree ->
  Spawnlib.spawn_return *)
let spawn_loop iotreep iotree =
  let spawnf (ic, oc) (ioatreep, ioatree) =
    let treep = Iotree.iotree_of_ioatree ioatreep in
    let tree = Iotree.iotree_of_ioatree ioatree in
    let p = Iotree.map_tree treep in
    let f = Iotree.map_tree tree in
    Sklstream.to_channel
      (Sklstream.loop p f (Sklstream.of_channel ic)) oc in
  try
    let port = Spawnlib.spawn_sock spawnf (iotreep, iotree) in
    Log.say (Log.Log 1) "Spawn loop success %d" port;
    Spawnd_types.Spawn_success
      { Spawnd_types.worker_port = port; }
  with IO.IOError msg ->
    Log.say Log.Err "IO error in Slave.spawn_loop: %s" msg;
    Spawnd_types.Spawn_failure
;;

(* val spawn_pipe : ('a, 'b) Iotree.ioatree -> ('b, 'c) Iotree.ioatree ->
  Spawnlib.spawn_return *)
let spawn_pipe iotree1 iotree2 =
  let spawnf (ic, oc) (iotree1, iotree2) =
    let os1 = to_tree iotree1 (Sklstream.of_channel ic) in
    let os2 = to_tree iotree2 os1 in
    Sklstream.to_channel os2 oc in
  try
    let port = Spawnlib.spawn_sock spawnf (iotree1, iotree2) in
    Log.say (Log.Log 1) "Spawn pipe success %d" port;
    Spawnd_types.Spawn_success
      { Spawnd_types.worker_port = port; }
  with IO.IOError msg ->
    Log.say Log.Err "IO error in Slave.spawn_pipe: %s" msg;
    Spawnd_types.Spawn_failure
;;

(* val slave_loop : ic * oc -> node -> unit *)
let rec slave_loop io node =
  let req_handler req =
    match req with
    | Spawnd_types.Spawn_atom (params, id) ->
      spawn_atom node.exe_path (params, id)
    | Spawnd_types.Spawn_loop (tree1, tree2) ->
      spawn_loop tree1 tree2
    | Spawnd_types.Spawn_pipe (tree1, tree2) ->
      spawn_pipe tree1 tree2 in
  Spawnlib.slave_loop io req_handler
;;

(* val do_dynamic_link : unit -> unit *)
let do_dynamic_link () =
  let load s =
    if not Dynlink.is_native
    then
      try Dynlink.loadfile s with Dynlink.Error e ->
        Log.say Log.Warn
          "Warning: dynamic link error, %s"
          (Dynlink.error_message e)
    else Log.say Log.Warn
      "Warning: %S was not loaded, running in opt"
      s in
  let rec chomp s =
    if s = "" then s else
    if s.[0] = ' ' || s.[0] = '\t'
    then chomp (String.sub s 1 (String.length s - 1))
    else s in
  let rec aux s =
    let index =
      try Some (String.index s ';') with _ -> None in
    match index with
    | Some i ->
      let fn = String.sub s 0 i in
      load (chomp fn);
      aux (String.sub s (i + 1) (String.length s - i - 1))
    | None ->
      let s = chomp s in
      if s = "" then () else load s
  in
  Dynlink.allow_unsafe_modules true;
  aux (try Sys.getenv "SKLML_DLINK" with Not_found -> "")
;;

(* val start : node -> unit *)
let start node =
  let sl_packet = { Spawnd_types.color = node.color; } in
  (if node.link then do_dynamic_link else ignore) ();
  try
    let (_, oc) as io = IO.connect node.master in
    IO.to_channel oc sl_packet true;
    slave_loop io node
  with
    IO.IOError msg -> Log.say Log.Err "IO error, %s" msg; exit 1
;;
